/*
* This file has primitives for partitioning loop iterations
*/
#include <stdio.h>
#include <math.h>
#include "parti.h"
#include "time.h"

#define TRUE 1
#define FALSE 0

/***********************************************************/
/* IN:                                                     */
/* ttable : translation table for lhs array indices        */
/* color : array returned by partitioner describing partition */
/* my_size : size of lhs indices                           */
/* my_loop_size: the  size of loop iteration               */
/* ref_per_iter: no of different refs considered per iter  */
/* OUT:                                                    */
/* loop_color : array that describes the loop partition    */
/***********************************************************/

/*================================================================*/
dref_rig(ttable, ind_referred, my_loop_size, ref_per_iter, ripa)
TTABLE *ttable;
int ref_per_iter, my_loop_size;
int *ind_referred;
int *ripa;
{
#ifdef DEBUG 
   int per_proc[128];
   int iwork[128];
   for(i=0; i<128; i++) per_proc[i] = 0;
#endif  
   FILE *fp, fp2;

   int i;
   int *local;


  if(ref_per_iter == 0)
     printf("something worng here! dref_rig \n"), exit(0);

  local     = (int *) malloc(sizeof(int)*my_loop_size*ref_per_iter);

  dereference(ttable, ind_referred, local, ripa, my_loop_size*ref_per_iter);

  free(local);
}

/*================================================================*/

iteration_partitioner_old(ripa, my_loop_size, ref_per_iter, color)
int *ripa, ref_per_iter, my_loop_size;
int *color;
{

   int i, j, k;
   int found, icount;
   int cur_proc, n_diff_proc, *count, *proc;
   int max_proc, max_count;
   int  n_off_proc;
   int rand_gen();
   double drand48();

  proc     = (int *) malloc(sizeof(int)*ref_per_iter);
  count    = (int *) malloc(sizeof(int)*ref_per_iter);

  srand48(0);
  if (ref_per_iter == 1) {
    for(i=0; i<my_loop_size; i++) {
      color[i] = ripa[i];
    }
  }
  icount = 0;
  if (ref_per_iter > 1)  {
    for(i=0; i<my_loop_size; i++) {
     for(j=0; j<ref_per_iter; j++) {
        count[j] = 0;
     }
     n_diff_proc = 0;
     max_count   = -1;
     for (j=0; j<ref_per_iter; j++) {
        cur_proc = ripa[icount++];
        found = FALSE;
        for(k=0; k<n_diff_proc; k++) {
          if(proc[k] == cur_proc) {
            count[k]++;
            if(max_count < count[k]) {
              max_count = count[k];
              max_proc = k;
              found    = TRUE;
              break;
            }
          }
        }
        if (!found) {
          proc[n_diff_proc] = cur_proc; 
          count[n_diff_proc]++;
          if(max_count < count[n_diff_proc]) {
            max_count = count[n_diff_proc];
            max_proc  = n_diff_proc;
            n_diff_proc++; 
          }
        }
     }

     if (max_count == 1) {
        color[i] = proc[(int) (drand48()*ref_per_iter)];
     }
     else {
         color[i] = proc[max_proc];
     }
#ifdef DEBUG 
     if (proc_assigned[i] >= 31 ) printf("error: in %d proc[0] %d proc[1] %d\n",
            i, proc[0], proc[1]);
     per_proc[proc_assigned[i]]++;
#endif
   }
  }

#ifdef DEBUG 
   printf("Loop partition \n");
   gisum(per_proc, PARTI_numnodes, iwork);
   if (PARTI_mynode == 0) {
     printf("number of edges \n");
     for(i=0; i<PARTI_numnodes; i++) {
      printf("[%d] %d\n", i, per_proc[i]);
     }
   }
#endif
   free(count);
   free(proc);
/*
   free_sched(lsched);
*/
}
/*================================================================*/

iteration_partitioner(ripa, my_loop_size, ref_per_iter, itable)
int *ripa, ref_per_iter, my_loop_size;
TTABLE **itable;
{

   int i, j, k;
   int found;
   int cur_proc, n_diff_proc, *count, *proc;
   int *color;
   int max_proc, max_count;
   int  n_off_proc;
   int rand_gen();
   double drand48();
   FILE *fp;
   TTABLE * init_ttable_with_proc();

  color     = (int *) malloc(sizeof(int)*my_loop_size);
  proc     = (int *) malloc(sizeof(int)*ref_per_iter);
  count    = (int *) malloc(sizeof(int)*ref_per_iter);

  srand48(0);
  if (ref_per_iter == 1) {
    for(i=0; i<my_loop_size; i++) {
      color[i] = ripa[i];
    }
  }
  if (ref_per_iter > 1)  {
    for(i=0; i<my_loop_size; i++) {
     for(j=0; j<ref_per_iter; j++) {
        count[j] = 0;
     }
     n_diff_proc = 0;
     max_count   = -1;
     for (j=0; j<ref_per_iter; j++) {
        cur_proc = ripa[i*ref_per_iter+j];
        found = FALSE;
        for(k=0; k<n_diff_proc; k++) {
          if(proc[k] == cur_proc) {
            count[k]++;
            if(max_count < count[k]) {
              max_count = count[k];
              max_proc = k;
              found    = TRUE;
              break;
            }
          }
        }
        if (!found) {
          proc[n_diff_proc] = cur_proc; 
          count[n_diff_proc]++;
          if(max_count < count[n_diff_proc]) {
            max_count = count[n_diff_proc];
            max_proc  = n_diff_proc;
            n_diff_proc++; 
          }
        }
     }

     if (max_count == 1) {
        color[i] = proc[(int) (drand48()*ref_per_iter)];
     }
     else {
         color[i] = proc[max_proc];
     }
#ifdef DEBUG 
     if (proc_assigned[i] >= 31 ) printf("error: in %d proc[0] %d proc[1] %d\n",
            i, proc[0], proc[1]);
     per_proc[proc_assigned[i]]++;
#endif
   }
  }

/*
  if (PARTI_mynode == 0 && ref_per_iter == 2) {
  fp = fopen("loop_iter", "w");
    for(i=0; i<my_loop_size; i++) {
		fprintf(fp,"  %d  ", i); 
		for(j=0; j<ref_per_iter; j++) {
			fprintf(fp,"  %d  ", ripa[j*my_loop_size+i]); 
        }
		fprintf(fp," color  %d \n  ", color[i]); 
    }
   fclose(fp);
  }
*/
  *itable = (TTABLE *) init_ttable_with_proc(1, color, my_loop_size);

#ifdef DEBUG 
   printf("Loop partition \n");
   gisum(per_proc, PARTI_numnodes, iwork);
   if (PARTI_mynode == 0) {
     printf("number of edges \n");
     for(i=0; i<PARTI_numnodes; i++) {
      printf("[%d] %d\n", i, per_proc[i]);
     }
   }
#endif
   free(count);
   free(proc);
   free(color);
/*
   free_sched(lsched);
*/
}

int
rand_gen(max)
int max;
{
  int rtnval;
  double drand48();
  
  rtnval = (int) (drand48()*max);
  return rtnval;
}


iter_part(ttable, ind_referred, my_loop_size, ref_per_iter,  color)
TTABLE *ttable;
int ref_per_iter, my_loop_size;
int *ind_referred;
int *color;
{
   FILE *fp, fp2;

   int i, j, k;
   int *local;
   int *ripa;
   int found;
   int cur_proc, n_diff_proc, *count, *proc;
   int max_proc, max_count;
   int  n_off_proc;
   int rand_gen();
   double drand48();


   if(ref_per_iter == 0)
     printf("something worng here! iter_part \n"), exit(0);

  ripa      = (int *) malloc(sizeof(int)*my_loop_size*ref_per_iter);
  local     = (int *) malloc(sizeof(int)*my_loop_size*ref_per_iter);

  dereference(ttable, ind_referred, local, ripa, my_loop_size*ref_per_iter);

  free(local);

  /* assign a loop iteration to a processor which has maximum 
     local references */

  proc     = (int *) malloc(sizeof(int)*ref_per_iter);
  count    = (int *) malloc(sizeof(int)*ref_per_iter);

  srand48(0);
  if (ref_per_iter == 1) {
    for(i=0; i<my_loop_size; i++) {
      color[i] = ripa[i];
    }
  }
  if (ref_per_iter > 1)  {
    for(i=0; i<my_loop_size; i++) {
     for(j=0; j<ref_per_iter; j++) {
        count[j] = 0;
     }
     n_diff_proc = 0;
     max_count   = -1;
     for (j=0; j<ref_per_iter; j++) {
        cur_proc = ripa[j*my_loop_size+i];
        found = FALSE;
        for(k=0; k<n_diff_proc; k++) {
          if(proc[k] == cur_proc) {
            count[k]++;
            if(max_count < count[k]) {
              max_count = count[k];
              max_proc = k;
              found    = TRUE;
              break;
            }
          }
        }
        if (!found) {
          proc[n_diff_proc] = cur_proc; 
          count[n_diff_proc]++;
          if(max_count < count[n_diff_proc]) {
            max_count = count[n_diff_proc];
            max_proc  = n_diff_proc;
            n_diff_proc++; 
          }
        }
     }

     if (max_count == 1) {
        color[i] = proc[(int) (drand48()*ref_per_iter)];
     }
     else {
         color[i] = proc[max_proc];
     }

#ifdef DEBUG 
     if (proc_assigned[i] >= 31 ) printf("error: in %d proc[0] %d proc[1] %d\n",
            i, proc[0], proc[1]);
     per_proc[proc_assigned[i]]++;
#endif
   }
  }

#ifdef DEBUG 
   printf("Loop partition \n");
   gisum(per_proc, PARTI_numnodes, iwork);
   if (PARTI_mynode == 0) {
     printf("number of edges \n");
     for(i=0; i<PARTI_numnodes; i++) {
      printf("[%d] %d\n", i, per_proc[i]);
     }
   }
#endif
   free(count);
   free(proc);
   free(ripa);
}

